/******************************************************************************************
 * Projeto Sensor Alagamento
 * 
 * Dispositivo para detectar alagamentos de vias pblicas e outras com aviso audio visual e
 * via rede
 * 
 * Desenvolvido por: Eng. Marcio Jose Soares
 * 
 * Plataforma: ESP8266 (Lolim D1 R1 Wemos)
 * Compilador: IDE Arduino 1.8.15 Linux
 * 
 * Pinos:
 * D3 - Entrada do sensor
 * D4 - saida sirene
 * D5 - saida LP1
 * D6 - saida LP2
 * 
 * ***************************************************************************************
 *  Ultimas alteracoes:
 *  em 02/02/2021:
 *    - criado esse projeto
******************************************************************************************/

/******************************************************************************************
* Arquivos includos no modulo
******************************************************************************************/
#include <Arduino.h>
#include "ESP8266WiFi.h"
#include <WiFiUdp.h>

/******************************************************************************************
 * Configuraçao para rede Wifi
 * 
 * Formato do arquivo rede.h
 * 
 * #ifndef STASSID
 * #define STASSID "nome da rede"
 * #define STAPSK  "senha da rede"
 * #endif
******************************************************************************************/
#include "rede.h"         //contem informaçao da rede
                          
/******************************************************************************************
 * Pinos de I/O
******************************************************************************************/
#define LED_LIVE  D2
#define SENSOR    D3      //GPIO5
#define SIREN     D4      //GPIO4
#define LAMP1     D5      //GPIO14
#define LAMP2     D6      //GPIO12

/******************************************************************************************
 * Macros para pinos de I/O
******************************************************************************************/
#define SIRENon   (digitalWrite(SIREN, HIGH))
#define SIRENoff  (digitalWrite(SIREN, LOW))
#define LAMP1on   (digitalWrite(LAMP1,HIGH))
#define LAMP1off  (digitalWrite(LAMP1,LOW))
#define LAMP2on   (digitalWrite(LAMP2,HIGH))
#define LAMP2off  (digitalWrite(LAMP2,LOW))
#define LEDon     (digitalWrite(LED_LIVE,HIGH))
#define LEDoff    (digitalWrite(LED_LIVE,LOW))

/******************************************************************************************
 * Constantes
******************************************************************************************/
const char * ssid = STASSID;                // nome da rede wifi
const char * pass = STAPSK;                 // sennha

const char* ntpServerName = "b.ntp.br";
const int NTP_PACKET_SIZE = 48;             // time stamp do NTP nos primeiros 48 bytes
const int UDP_PACKET_SIZE = 128;            // tamanho maximo do pacote UDP enviado

const long  gmtOffset_sec = 0;
const int   daylightOffset_sec = 3600;

const unsigned char _7segdata[10] = {0x3f, 0x03, 0x6d, 0x67, 0x53, 0x76, 0x7e, 0x23, 0x7f, 0x77};

const char* IDEQUIPO = "02";    // cada equipamento tem seu ID 
const char* RGEQUIPO = "SUL";   // "NORTE";"SUL";"LESTE";"OESTE";"CENTRO" - escolha apenas uma!

/******************************************************************************************
 * Variveis
******************************************************************************************/
IPAddress timeServerIP;                     // servidor NTP
IPAddress centralIP(0, 0, 0, 0);       		// ip da central que receber o alerta
int centralPORT = 2400;                     // porta de comunicaçao UDP com a central

WiFiUDP udp;                                // instancia UDP 
unsigned int localPort = 2390;              // porta para bind UDP - msgs enviadas pela central

byte packetBuffer[NTP_PACKET_SIZE];        // buffer para pacotes UDP

int val = 0;
int state_led = LOW;
int estado_lamps = LOW;
int alerta = 0;

volatile int time500ms = 0;
volatile int time1s = 0;
volatile int time2s = 0;
volatile int time5s = 0;

char mytime[UDP_PACKET_SIZE];

/******************************************************************************************
 * Declara funçoes do modulo
******************************************************************************************/
void getmyNTP(void);
void sendAlertaUDP(void);
void sendNTPpacket(IPAddress& address);

/******************************************************************************************
 * Setup
******************************************************************************************/
void setup() {
    Serial.begin(115200);
    Serial.println();
    Serial.println();

    pinMode(SENSOR, INPUT_PULLUP);
    pinMode(SIREN, OUTPUT);
    pinMode(LAMP1, OUTPUT);
    pinMode(LAMP2, OUTPUT);
    pinMode(LED_LIVE, OUTPUT);

    SIRENoff;
    LAMP1off;
    LAMP2off;
    LEDoff;
    
    Serial.print("Conectando a ");
    Serial.println(ssid);
    WiFi.mode(WIFI_STA);
    WiFi.begin(ssid, pass);

    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
    }
    Serial.println("");

    Serial.println("WiFi conectado");
    Serial.println("Endereco IP: ");
    Serial.println(WiFi.localIP());

    Serial.println("Iniciando UDP");
    udp.begin(localPort);
    Serial.print("Porta Local: ");
    Serial.println(udp.localPort());
    
}

/******************************************************************************************
 * Loop - funcao principal
 * 
 * Entradas - nenhuma
 * Saidas   - nenhuma
******************************************************************************************/
void loop() {

    delay(10);
    time500ms++;                                            //
    if(time500ms == 50){
        time500ms = 0;
        if(state_led == LOW){
            LEDon;
            state_led = HIGH;
        }
        else{
            LEDoff;
            state_led = LOW;
        }

         if(alerta == 1){                               //foi ativado o alerta?
             time1s++;                                  //altera estado das lampadas a cada 1s
             if(time1s == 2){                           //altera?
                 time1s = 0;                            //se sim zera para proxima contagem
                 if(estado_lamps == LOW){               //verifica estado e troca
                     LAMP1on;
                     LAMP2off;
                     estado_lamps = HIGH;
                 }
                 else{
                     LAMP1off;
                     LAMP2on;
                     estado_lamps = LOW;
                 }
             }
             SIRENon;                                  //liga sirene
         }
         else{                                         //alerta desligado
             LAMP1off;                                 //mantem tudo desligado
             LAMP2off;
             SIRENoff;
             estado_lamps = LOW;
         }

         time5s++;                                      //incrementa o tempo
         if(time5s == 10){                              //verifica o sensor a cada 5 segundos
             time5s = 0;
             val = digitalRead(SENSOR);                          //le o sensor
             if(val == LOW){                                     //sensor esta ativo?!
                 alerta = 1;                                     //emite o alerta
                 WiFi.hostByName(ntpServerName, timeServerIP);   //pega servidor aleatorio
                 sendNTPpacket(timeServerIP);                    // envia pacote para servidor NTP
                 //delay(1000);                                    //aguarda
                 getmyNTP();                                     //coleta data e hora do servidor 
                 sendAlertaUDP();                                //envia alerta para central
             }
             else{
                 alerta = 0;                                     //desliga alerta!
             }
        }    
    }
}

/******************************************************************************************
 * Funçao para enviar pacote UDP para a central
 * 
 * Entradas - nenhuma
 * Saidas   - nenhuma
******************************************************************************************/
void sendAlertaUDP(void){

    int i;
    char mypacket[UDP_PACKET_SIZE+2];
    
    for(i=0;i<sizeof(mytime);i++){             //coloca os dados
        mypacket[i] = mytime[i];               //insere a hora do evento
    }
    
    mypacket[i] = '\0';                        //finaliza string
    
    udp.beginPacket(centralIP, centralPORT);   //inicia comunicacao
    udp.write(mypacket, strlen(mypacket));     //envia pacote
    udp.endPacket();                           //encerra comunicacao
        
}

/******************************************************************************************
 * Funçao para pegar pacote de resposta do servidor NTP
 * 
 * Entradas - nenhuma
 * Saidas   - nenhuma
******************************************************************************************/
void getmyNTP(void){

    int i;
    char p[10];

    int myhour = 0;
    int mymin = 0;
    int mysec = 0;

    strcpy(mytime, "Alagamento na regiao:");
    strcat(mytime, RGEQUIPO);
    strcat(mytime, " Equipamento:");
    strcat(mytime, IDEQUIPO);
    //strcat(mytime, "IP:");
    //strcat(mytime,(const char*)WiFi.localIP());
    strcat(mytime, " hora->");
    
    int cb = udp.parsePacket();
    if (!cb) {
       Serial.println("sem pacote ainda");
    }
    else {
       Serial.print("pacote recebido, tamanho=");
       Serial.println(cb);
    
        // Pacote recebido, leia dados do mesmo
        udp.read(packetBuffer, NTP_PACKET_SIZE);    // le pacote do buffer
        //o timestampe começa no byte 40 do pacote recebido e sao 4 bytes,
        //or duas words, long. Primeiro extraindo as duas words:

        unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
        unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
     
        // combine os 4 bytes (2 words) em um inteiro longo
        // este  o NTP time (segundos desde 1 de Janjeiro de 1900):
        unsigned long secsSince1900 = highWord << 16 | lowWord;
        Serial.print("Segundos desde 1 de Janeiro de 1990 = ");
        Serial.println(secsSince1900);

        // Agora converte:
        Serial.print("Unix time = ");
        // Unix time coemça em  1 de Janeiro de 1970. Em segundos, sao 2208988800:
        const unsigned long seventyYears = 2208988800UL;
        // Tira 70 anos:
        unsigned long epoch = secsSince1900 - seventyYears;
        // mostra Unix time:
        Serial.println(epoch);

        //agora eh hora de acertar o relogio...
        epoch = epoch - (3600 * 3);                 //acerta o GMT -3 
    
        // mostra a hora, minutos e segundos:
        Serial.print("O UTC time eh ");             // UTC is the time at Greenwich Meridian (GMT)
        myhour = (epoch  % 86400L) / 3600;
        mymin = (epoch % 3600) / 60;
        mysec = (epoch % 60);
        
        Serial.print(myhour);                       // print the hour (86400 equals secs per day)
        Serial.print(':');

        if (mymin < 10) {                           //colaca zero se necessrio
            Serial.print('0');
        }
        Serial.print(mymin);             //minuto
        Serial.print(':');

        if (mysec < 10) {
            Serial.print('0');
        }
        
        Serial.print(epoch % 60); // print the second
        Serial.println(" - GMT=-3");

        itoa(myhour, p, 10);
        strcat(mytime,(const char *)p);
        strcat(mytime,":");
        if (mymin < 10) {                           //colaca zero se necessrio
            strcat(mytime,"0");
        }
        itoa(mymin, p, 10);
        strcat(mytime,(const char*)p);
        strcat(mytime,":");
        
        if (mysec < 10) {
            strcat(mytime,"0");
        }
        itoa(mysec, p, 10);
        strcat(mytime,(const char*)p);
        
        Serial.println(mytime);
        
    }
}

/******************************************************************************************
 * Funçao para enviar pacote NTP para servidor NTP
 * 
 * Entradas - endereço IP
 * Saidas   - nenhuma
******************************************************************************************/
void sendNTPpacket(IPAddress& address){
  
    Serial.println("\n\nEnviando pacote NTP...");
    memset(packetBuffer, 0, NTP_PACKET_SIZE);
  
    packetBuffer[0] = 0b11100011;   // LI, Version, Mode
    packetBuffer[1] = 0;            // Stratum, or type of clock
    packetBuffer[2] = 6;            // Polling Interval
    packetBuffer[3] = 0xEC;         // Peer Clock Precision
 
    packetBuffer[12]  = 49;
    packetBuffer[13]  = 0x4E;
    packetBuffer[14]  = 49;
    packetBuffer[15]  = 52;
  
  
    udp.beginPacket(address, 123); //NTP requests are to port 123
    udp.write(packetBuffer, NTP_PACKET_SIZE);
    udp.endPacket();
}
